"""This cube handles classification schemes.

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from yams.buildobjs import (EntityType, RelationType,
                            SubjectRelation, ObjectRelation, String)

from cubicweb.schema import RQLConstraint, ERQLExpression


_ = unicode

class Classification(EntityType):
    """classification schemes are used by users to classify entities.
    """
    __permissions__ = {
        'read' : ('managers', 'users', 'guests'),
        'add' : ('managers',),
        'delete' : ('managers',),
        'update' : ('managers',),
        }
    name = String(required=True, fulltextindexed=True, unique=True,
                  maxsize=128)
    classifies = SubjectRelation('CWEType',
                     # see relation type docstring
                     constraints = [RQLConstraint('RDEF to_entity O,'
                                                  'RDEF relation_type R,'
                                                  'R name "applied_to"',
                     msg="Classification is trying to classifies an EntityType "
                         "without applied_to relation")])


class classifies(RelationType):
    """entity types classified by the classification. Only entity type
    supporting the applied_to relation can be selectioned
    """


class Keyword(EntityType):
    """A keyword is like a tag but is application specific
    and used to define a classification scheme
    """
    __permissions__ = {
        'read' : ('managers', 'users', 'guests'),
        'add' : ('managers', 'users'),
        'delete' : ('managers',),
        'update' : ('managers',),
        }
    name = String(required=True, fulltextindexed=True, indexed=True, maxsize=128)

    subkeyword_of = SubjectRelation('Keyword', cardinality='?*',
                                    description=_('which keyword (if any) this keyword specializes'),
                                    # if no included_in set, it'll be automatically added by a hook
                                    constraints=[RQLConstraint('NOT S included_in CS1 OR EXISTS(S included_in CS2, O included_in CS2)')])
    descendant_of = SubjectRelation('Keyword')
    included_in = SubjectRelation('Classification', cardinality='1*')


class CodeKeyword(Keyword):
    """A CodeKeyword is a keyword with a code and a name
    """
    __specializes_schema__ = True
    code = String(required=True, fulltextindexed=True, indexed=True, maxsize=128)

class subkeyword_of(RelationType):
    """a keyword can specialize another keyword"""


class included_in(RelationType):
    """a keyword is included in a classification scheme"""
    inlined = True


# define in parent application which entity types may be linked to a keyword
# by the applied_to relation

class applied_to(RelationType):
    """tagged objects"""
    fulltext_container = 'object'
    constraints = [RQLConstraint('S included_in CS, O is ET, CS classifies ET')]



"""entity classes for classification schemes entities

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb.mixins import MI_REL_TRIGGERS, TreeMixIn
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.interfaces import ITree


class Classification(AnyEntity):
    __regid__ = 'Classification'
    fetch_attrs, fetch_order = fetch_config(['name'])
    __implements__ = AnyEntity.__implements__ + (ITree,)

    def root(self):
        """returns the root object"""
        return None

    def parent(self):
        """returns the parent entity"""
        return None

    def iterparents(self):
        """returns parent entities"""
        yield self

    def children(self, entities=True):
        """returns the item's children"""
        return self.related('included_in', 'object', entities=entities)

    def children_rql(self):
        """XXX returns RQL to get children"""
        return self.related_rql('included_in', 'object')

    def is_leaf(self):
        """returns true if this node as no child"""
        return bool(self.children())

    def is_root(self):
        """returns true if this node has no parent"""
        return True

    def first_level_keywords(self):
        return self.req.execute('Any K,N ORDERBY N WHERE K included_in C, '
                                'NOT K subkeyword_of KK, K name N, '
                                'C eid %(x)s', {'x': self.eid})


class Keyword(TreeMixIn, AnyEntity):
    __regid__ = 'Keyword'
    fetch_attrs, fetch_order = fetch_config(['name'])
    __implements__ = AnyEntity.__implements__ + (ITree,)

    tree_attribute = 'subkeyword_of'

    @property
    def classification(self):
        if self.included_in:
            return self.included_in[0]
        return None

    def parent(self):
        """IBreadcrumbs implementation"""
        if self.subkeyword_of:
            return self.subkeyword_of[0]
        return self.classification

    def iterparents(self):
        """returns parent keyword entities
        """
        if self.subkeyword_of:
            parent = self.subkeyword_of[0]
            while parent is not None:
                yield parent
                if parent.subkeyword_of:
                    parent = parent.subkeyword_of[0]
                else:
                    parent = None

    def children(self, entities=True):
        """returns the item's children

        we have only one direct child by ``subkeyword_of`` relation"""
        assert 1 == len(self.reverse_subkeyword_of)
        return iter(self.reverse_subkeyword_of)

    def iterchildren(self):
        """returns children entities"""
        if self.reverse_subkeyword_of:
            child = self.reverse_subkeyword_of[0]
            while child is not None:
                yield child
                if child.reverse_subkeyword_of:
                    child = child.reverse_subkeyword_of[0]
                else:
                    child = None

    def is_leaf(self):
        if self.reverse_subkeyword_of:
            return False
        return True

    def children_rql(self):
        return 'Any K WHERE  K subkeyword_of X, X eid %(x)s'

    """
    # FIXME unittest
    def subkeywords(self, recursive=True):
        rset = self.req.execute(self.children_rql(), {'x': self.eid})
        subentities = list(rset.entities())
        if recursive:
            for entity in subentities[:]:
                subentities.extend(entity.subkeywords(recursive=True))
        return subentities
    """

class CodeKeyword(Keyword):
    __regid__ = 'CodeKeyword'
    rest_attr = 'code'
    fetch_attrs, fetch_order = fetch_config(['code','name'])

    def dc_title(self):
        return u'%s - %s' % (self.code, self.name)


class ClassifiableMixIn(object):
    """mixin automatically plugged to entity types supporting the applied_to
    relation
    """
    def classification_keywords(self, name):
        """return keywords of the given classification linked to this entity"""
        return self.req.execute('Any K, KN WHERE K applied_to X, X eid %(x)s, '
                                'K name KN, K included_in CS, CS name %(name)s',
                                {'x': self.eid, 'name': name}, 'x')

MI_REL_TRIGGERS[('applied_to', 'object')] = ClassifiableMixIn



from logilab.common.testlib import unittest_main
from cubicweb.devtools.testlib import AutomaticWebTest

class AutomaticWebTest(AutomaticWebTest):

    ignored_relations = set(('descendant_of',))

    def to_test_etypes(self):
        # add cwgroup to test keyword components on entity on which they can be applied
        return set(('Classification', 'Keyword', 'CWGroup'))

    def list_startup_views(self):
        return ()

if __name__ == '__main__':
    unittest_main()



# pylint: disable-msg=W0622
"""cubicweb-classification-schemes packaging information"""

modname = 'keyword'
distname = "cubicweb-keyword"

numversion = (1, 6, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
copyright = '''Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr'''

author = "Logilab"
author_email = "contact@logilab.fr"
web = 'http://www.cubicweb.org/project/%s' % distname

short_desc = "keyword component for the Cubicweb framework"
long_desc = """Summary
-------
The `keyword` cube provides classification by using hierarchies of keywords to
classify content.

There is two types of keywords:

- `Keyword` which contains a description,
- `CodeKeyword` which contains the keyword description and the
  associated code.

In order to link an entity to a keyword, you have to add a relation
`applied_to` in the schema.

Each keyword has the `subkeyword_of` relation definition. This allows
to navigate in the classification without a Modified Preorder Tree
Traversal representation of the data.

Some methods are defined in order to get parents and children or get
the status of a keyword (leaf or root).

See also cubicweb-tag as another (simpler) way to classify content.
"""

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]

__depends_cubes__ = {}
__depends__ = {'cubicweb': '>= 3.6.0'}
__use__ = tuple(__depends_cubes__)

from os import listdir
from os.path import join

CUBES_DIR = join('share', 'cubicweb', 'cubes')
try:
    data_files = [
        [join(CUBES_DIR, 'keyword'),
         [fname for fname in listdir('.')
          if fname.endswith('.py') and fname != 'setup.py']],
        [join(CUBES_DIR, 'keyword', 'data'),
         [join('data', fname) for fname in listdir('data')]],
        [join(CUBES_DIR, 'keyword', 'i18n'),
         [join('i18n', fname) for fname in listdir('i18n')]],
        [join(CUBES_DIR, 'keyword', 'migration'),
         [join('migration', fname) for fname in listdir('migration')]],
        ]
except OSError:
    # we are in an installed directory
    pass
